#!python
#
# SRTMCompare.py
#
# Script to compare regions of 3" SRTM tiles

import os
import sys
import array
import re
import zipfile
import ctypes

# Source and destination filenames and data structures
SRTMExt		= ".hgt"
SRTMPath1	= None
SRTMPath2	= None
RawFileName	= "SRTMCompareImg"
SRTMX		= 1201
SRTMY		= 1201
DstPath		= ".\\";
BakExt		= ".bak"
RawExt		= ".raw"
TmpExt		= ".tmp"
ZipExt		= ".zip"
Scale		= float( 192 )
Start		= [ -8, 49 ]
Finish		= [ +2, 61 ]
UKLimits	= [ "N51W008", "N52W008", "N52W007", "N53W008", "N53W007", "N54W008", "N54W007", "N55W008" ]
# Value flagging a void or an invalid height
InvHeight	= -32768;


def main():
	global Start, Finish, SRTMPath1, SRTMPath2, RawFileName, Scale
	SRTMMin		= 0
	SRTMMax		= 0
	GlobTiles	= 0
	GlobChgs	= 0
	GlobVoids	= 0
	GlobFilld	= 0
	GlobMinChg	= 32767
	GlobMaxChg	= -32767
	GlobAveChg	= 0
	Debug		= False
	Full 		= False
	Grade 		= False
	Raw 		= False
	UK	 		= False
	for opt in sys.argv[1:]:
		if re.compile( "^[-/]F$", re.IGNORECASE ).match( opt ):
			Raw		= True
			Full	= True
			print	"RAW image will be created as visual check ..."
			print	"Using full colour saturation ..."
		elif re.compile( "^[-/]G$", re.IGNORECASE ).match( opt ):
			Raw		= True
			Grade	= True
			print	"RAW image will be created as visual check ..."
			print	"Using graded colour saturation ..."
		elif re.compile( "^[-/]R$", re.IGNORECASE ).match( opt ):
			Raw		= True
			print	"RAW image will be created as visual check ..."
		elif re.compile( "^[-/]D$", re.IGNORECASE ).match( opt ):
			Debug	= True
			Start	= [ -7, 55 ]
			Finish	= [ -6, 56 ]
			print	"Debugging ..."
		elif re.compile( "^[-/]S$", re.IGNORECASE ).match( opt ):
			Start	= [ -3, 59 ]
			Finish	= [ 0, 61 ]
			print	"Shetland only ..."
		elif re.compile( "^[-/]T$", re.IGNORECASE ).match( opt ):
			Start	= [ -7, 55 ]
			Finish	= [ 0, 56 ]
			print	"Testing only ..."
		elif re.compile( "^[-/]U$", re.IGNORECASE ).match( opt ):
			UK			= True
			Start		= [ -8, 49 ]
			Finish		= [ +2, 61 ]
			print	"Strict UK only ..."
		elif SRTMPath1 == None:
			SRTMPath1	= os.path.join( opt, "" )
			print		"First path: %s" % ( SRTMPath1 )
		elif SRTMPath2 == None:
			SRTMPath2	= os.path.join( opt, "" )
			print		"Second path: %s" % ( SRTMPath2 )

	TileRangeX  = range ( Start[0], Finish[0] )
	TileRangeY  = range ( Start[1], Finish[1] )

	# If creating RAW image, need to scan for range of heights
	if Raw:
		print "\nScanning SRTM tiles ..."
		for lat in TileRangeY:
			for lon in TileRangeX:
				TileName = "N%2d%c%03d" % ( lat, "W" if lon < 0 else "E", abs( lon ) )
				print "\nCurrent SRTM tile: %s" % ( TileName )

				# Read first SRTM tile
				SRTMFileName1 = SRTMPath1 + TileName + SRTMExt
				print "Reading first SRTM tile: %s" % ( SRTMFileName1 )
				try:
					bin = open( SRTMFileName1, "rb", 2*SRTMX*SRTMY )
					SRTMTile1 = array.array( "h", bin.read() )
					bin.close()
					if sys.byteorder == "little":
						SRTMTile1.byteswap()
				except IOError:
					try:
						bin = zipfile.ZipFile( SRTMFileName1 + ZipExt, "r" )
						SRTMTile1 = array.array( "h", bin.read(TileName + SRTMExt) )
						bin.close()
						if sys.byteorder == "little":
							SRTMTile1.byteswap()
					except IOError:
						print "Using blank starting tile for %s" % ( SRTMFileName1 )
						SRTMTile1 = array.array( "h", [0] * SRTMX * SRTMY )

				# Read second SRTM tile
				SRTMFileName2 = SRTMPath2 + TileName + SRTMExt
				print "Reading second SRTM tile: %s" % ( SRTMFileName2 )
				try:
					bin = open( SRTMFileName2, "rb", 2*SRTMX*SRTMY )
					SRTMTile2 = array.array( "h", bin.read() )
					bin.close()
					if sys.byteorder == "little":
						SRTMTile2.byteswap()
				except IOError:
					try:
						bin = zipfile.ZipFile( SRTMFileName2 + ZipExt, "r" )
						SRTMTile2 = array.array( "h", bin.read(TileName + SRTMExt) )
						bin.close()
						if sys.byteorder == "little":
							SRTMTile2.byteswap()
					except IOError:
						print "Using blank starting tile for %s" % ( SRTMFileName2 )
						SRTMTile2 = array.array( "h", [0] * SRTMX * SRTMY )

				if Debug:
					print "SRTMTile1 element size: %d, memory address:%d, number of elements:%d" % (SRTMTile1.itemsize, SRTMTile1.buffer_info()[0], SRTMTile1.buffer_info()[1] )
					print "SRTMTile2 element size: %d, memory address:%d, number of elements:%d" % (SRTMTile2.itemsize, SRTMTile2.buffer_info()[0], SRTMTile2.buffer_info()[1] )

				# Scan for maximum and minimum heights
				for sy in range ( 0, SRTMY ):
					for sx in range ( 0, SRTMX ):
						so		= (SRTMY - 1 - sy)*SRTMX + sx
						ht1		= SRTMTile1[so]
						ht2		= SRTMTile2[so]
						if ht1 != InvHeight:
							SRTMMin	= min( ht1, SRTMMin )
							SRTMMax	= max( ht1, SRTMMax )
						if ht2 != InvHeight:
							SRTMMin	= min( ht2, SRTMMin )
							SRTMMax	= max( ht2, SRTMMax )

		Scale /= (SRTMMax - SRTMMin)
		print "\nMinimum height:\t%d\nMaximum height:\t%d\nScale:\t%.3f" % ( SRTMMin, SRTMMax, Scale )

		# Create RAW image file for visual checking
		TmpFileName = DstPath + RawFileName + TmpExt
		if os.access( TmpFileName, os.F_OK ):
			print "Deleting existing temporary file: %s" % ( TmpFileName )
			os.remove( TmpFileName )
		print "Creating RAW image temporary file: %s" % ( TmpFileName )
		tmp = open( TmpFileName, "wb", 3*SRTMX*SRTMY )

		# Dimensions of the image
		RawDims	= ( (Finish[0] - Start[0]) * (SRTMX-1) + 1, (Finish[1] - Start[1]) * (SRTMY-1) + 1 )

	MinDiff = 32768
	MaxDiff	= 0
	AveDiff = float(0)
	DCount	= 0
	VCount	= 0

	for lat in TileRangeY:
		for lon in TileRangeX:

			TileName = "N%2d%c%03d" % ( lat, "W" if lon < 0 else "E", abs( lon ) )
			print "\nCurrent SRTM tile: %s" % ( TileName )

			# Read first SRTM file
			SRTMFileName1 = SRTMPath1 + TileName + SRTMExt
			print "Reading first SRTM tile: %s" % ( SRTMFileName1 )
			try:
				bin = open( SRTMFileName1, "rb", 2*SRTMX*SRTMY )
				SRTMTile1 = array.array( "h", bin.read() )
				bin.close()
				if sys.byteorder == "little":
					SRTMTile1.byteswap()
			except IOError:
				try:
					bin = zipfile.ZipFile( SRTMFileName1 + ZipExt, "r" )
					SRTMTile1 = array.array( "h", bin.read(TileName + SRTMExt) )
					bin.close()
					if sys.byteorder == "little":
						SRTMTile1.byteswap()
				except IOError:
					print "Using blank starting tile for %s" % ( SRTMFileName1 )
					SRTMTile1 = array.array( "h", [0] * (SRTMX * SRTMY) )

			# Read second SRTM file
			SRTMFileName2 = SRTMPath2 + TileName + SRTMExt
			print "Reading second SRTM tile: %s" % ( SRTMFileName2 )
			try:
				bin = open( SRTMFileName2, "rb", 2*SRTMX*SRTMY )
				SRTMTile2 = array.array( "h", bin.read() )
				bin.close()
				if sys.byteorder == "little":
					SRTMTile2.byteswap()
			except IOError:
				try:
					bin = zipfile.ZipFile( SRTMFileName2 + ZipExt, "r" )
					SRTMTile2 = array.array( "h", bin.read(TileName + SRTMExt) )
					bin.close()
					if sys.byteorder == "little":
						SRTMTile2.byteswap()
				except IOError:
					print "Using blank starting tile for %s" % ( SRTMFileName2 )
					SRTMTile2 = array.array( "h", [0] * (SRTMX * SRTMY) )

			# Scan to compare tiles
			for sy in range ( 0, SRTMY ):

				# Create blank raw image tile line
				if Raw:
					RawData = array.array( "B", [0]*3*SRTMX )
					if Debug:
						print "RawData element size: %d, memory address:%d, number of elements:%d" % (RawData.itemsize, RawData.buffer_info()[0], RawData.buffer_info()[1] )

				for sx in range ( 0, SRTMX ):
					so	= (SRTMY - 1 - sy)*SRTMX + sx
					ht1	= SRTMTile1[so]
					ht2	= SRTMTile2[so]
					dif	= abs(ht2 - ht1)

					if (dif != 0) and not (UK and (TileName in UKLimits)):
						if (ht1 != InvHeight) and (ht2 != InvHeight):
							DCount	+= 1
							AveDiff += dif
							MinDiff = min( dif, MinDiff )
							MaxDiff = max( dif, MaxDiff )
						else:
							VCount	+= 1
					if Raw:
						if (ht1 != InvHeight) and (ht2 != InvHeight):
							pi1 = int( round( Scale*(ht1 - SRTMMin) ) )
							pi2 = int( round( Scale*(ht2 - SRTMMin) ) )
							pi3 = int( round( Scale*((ht1+ht2)/2 - SRTMMin) ) )
							if Grade or Full:
								if Grade:
									sat	= dif
									if dif > 10:
										sat = 127
									if dif > 100:
										sat = 255
								elif Full:
									sat = 255
								if ht1 < ht2:
									pi1 = 0
									pi2 = sat
								elif ht1 > ht2:
									pi1 = sat
									pi2 = 0
								pi3 = 0
						else:
							pi1 = 0
							pi2 = 0
							pi3 = 255
							if ht1 == InvHeight:
								pi1 = 255
							if ht2 == InvHeight:
								pi2 = 255
						ro				= 3*sx
						RawData[ro]		= pi1
						RawData[ro + 1]	= pi2
						RawData[ro + 2]	= pi3

					if Debug:
						if Raw:
							print "sx: %4d, sy: %4d, so: %5d, ht1: %3d, ht2: %3d, ro: %5d, pi1: %3d, pi2: %3d, pi3: %3d, min: %5d, max: %3d" % ( sx, sy, so, ht1, ht2, ro, pi1, pi2, pi3, SRTMMin, SRTMMax )
						else:
							print "sx: %4d, sy: %4d, so: %5d, ht1: %3d, ht2: %3d, min: %5d, max: %3d" % ( sx, sy, so, ht1, ht2, SRTMMin, SRTMMax )

				if Raw:
					rx	= ( lon - Start[0] )*( SRTMX - 1 )
					ry	= ( lat - Start[1] )*( SRTMY - 1 ) + sy
					ro	= 3*( (RawDims[1] - 1 - ry)*RawDims[0] + rx )
					tmp.seek( ro )
					tmp.write( RawData )

	if Raw:
		tmp.close()

	print "\nCompared SRTM sets:\n\t%s\n\t%s\nMinimum Difference: %d\nMaximum Difference: %d\nAverage Difference: %.3f\nDifference Count: %d" % ( SRTMPath1, SRTMPath2, MinDiff, MaxDiff, AveDiff/DCount, DCount + VCount )

	if Raw:
		# Write RAW image file for visual checking
		print "\nCreating RAW image file %s as visual check\nIts dimensions will be %d x %d in 3 byte RGB" % ( RawFileName, RawDims[0], RawDims[1] )
		BakFileName = DstPath + RawFileName + BakExt
		RawFileName = DstPath + RawFileName + RawExt
		if os.access( BakFileName, os.F_OK ) and os.access( RawFileName, os.F_OK ):
			print "Deleting existing back up file: %s" % ( BakFileName )
			os.remove( BakFileName )
		if os.access( RawFileName, os.F_OK ):
			print "Backing up existing RAW file: %s" % ( RawFileName )
			os.rename( RawFileName, BakFileName )
		print "Finishing new RAW file: %s" % ( RawFileName )
		os.rename( TmpFileName, RawFileName )

main()
